# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''File containing unit tests for the BracketBase baseclass in
utils.py'''

import pytest
from fparser.two.utils import BracketBase, InternalError
from fparser.two.Fortran2003 import Name


def test_brackets():
    '''Test the bracketsbase match method with no content within the
    brackets.

    '''
    for lhs, rhs in [("(", ")"), (" ( ", " ) "), ("[", "]"), ("A", "A"),
                     ("([", "])"), ("{{[(", ")]}}"),
                     ("  {  {  [  (  ", " ) ] } } ")]:
        brackets = lhs + rhs
        input_text = brackets.replace(' ', '')
        result = BracketBase.match(brackets, None, input_text,
                                   require_cls=False)
        assert str(result) == "('{0}', None, '{1}')".format(
            lhs.replace(' ', ''), rhs.replace(' ', ''))


def test_input_too_short():
    '''Test the bracketsbase match method returns None when the input
    string is shorter than the brackets.

    '''
    brackets = "(())"
    input_text = "(H)"
    result = BracketBase.match(brackets, None, input_text,
                               require_cls=False)
    assert result is None


def test_cls():
    '''Test the bracketsbase match method with content within the brackets
    for both require_cls is False and True. require_cls indicates
    whether content should be expected within the brackets (and the
    match failing if not), or whether it is optional. The actual
    contents are passed on to the specified class ('Name' in this
    case) to match (or not), hence the name require_cls.

    '''
    for require in [False, True]:
        for lhs, rhs in [("(", ")"), (" ( ", " ) "), ("[", "]"), ("A", "A"),
                         ("([", "])"), ("{{[(", ")]}}")]:
            brackets = lhs + rhs
            input_text = lhs + "hello" + rhs
            result = BracketBase.match(brackets, Name, input_text,
                                       require_cls=require)
            assert str(result) == "('{0}', Name('hello'), '{1}')".format(
                lhs.replace(' ', ''), rhs.replace(' ', ''))


def test_brackets_error1():
    '''Test the bracketbase match method returns None if the brackets are
    invalid.'''

    for brackets in [None, "", "  ", "(", ")", "[[]"]:
        input_text = "()"
        result = BracketBase.match(brackets, None, input_text,
                                   require_cls=False)
        assert result is None


def test_brackets_error2():
    '''Test the bracketbase match method returns the brackets if class contents
    are expected but there are none.

    '''
    result = BracketBase.match("()", None, "()", require_cls=True)
    assert result is None


def test_brackets_error3():
    '''Test the bracketbase match method returns None if class contents
    are expected (as the default is require_cls=True) but there are
    none.

    '''
    result = BracketBase.match("()", None, "()")
    assert result is None


def test_tostr(monkeypatch):
    '''It is not possible to instantiate BracketBase directly so we create
    a class that uses BracketBase (Format_Specification) and then test
    the tostr() method from it. This test checks that an internal
    error is raised if the size of the internal items list is
    incorrect.

    '''
    from fparser.two.Fortran2003 import Format_Specification
    ast = Format_Specification("()")
    monkeypatch.setattr(ast, "items", [None])
    with pytest.raises(InternalError) as excinfo:
        _ = str(ast)
    assert "Class BracketBase method tostr()" in str(excinfo.value)
    assert "tostr() has '1' items, but expecting 3" in str(excinfo.value)


def test_tostr_invalid2(monkeypatch):
    '''It is not possible to instantiate BracketBase directly so we create
    a class that uses BracketBase (Format_Specification) and then test
    the tostr() method from it. This test checks that an internal
    error is raised if entry 0 of the internal items list is empty or
    None.

    '''
    from fparser.two.Fortran2003 import Format_Specification
    ast = Format_Specification("()")
    monkeypatch.setattr(ast, "items", [None, ast.items[1], ast.items[2]])
    with pytest.raises(InternalError) as excinfo:
        _ = str(ast)
    assert "Class BracketBase method tostr()" in str(excinfo.value)
    assert ("'Items' entry 0 should be a string containing the left hand "
            "bracket but it is empty or None") in str(excinfo.value)


def test_tostr_invalid3(monkeypatch):
    '''It is not possible to instantiate BracketBase directly so we create
    a class that uses BracketBase (Format_Specification) and then test
    the tostr() method from it. This test checks that an internal
    error is raised if entry 2 of the internal items list is empty or
    None.

    '''
    from fparser.two.Fortran2003 import Format_Specification
    ast = Format_Specification("()")
    monkeypatch.setattr(ast, "items", [ast.items[0], ast.items[1], None])
    with pytest.raises(InternalError) as excinfo:
        _ = str(ast)
    assert "Class BracketBase method tostr()" in str(excinfo.value)
    assert ("'Items' entry 2 should be a string containing the right hand "
            "bracket but it is empty or None") in str(excinfo.value)



# Copyright (c) 2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R1209 : This file tests the support for an
import statement.

'''

import pytest
from fparser.two.Fortran2003 import Import_Stmt
from fparser.two.utils import NoMatchError


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("example,result",
                         [("IMPORT", "IMPORT"),
                          ("iMpOrT", "IMPORT"),
                          ("  IMPORT  ", "IMPORT"),
                          ("IMPORT name1", "IMPORT :: name1"),
                          ("IMPORT name1, name2", "IMPORT :: name1, name2"),
                          ("IMPORT :: name1", "IMPORT :: name1"),
                          ("IMPORT :: name1, name2", "IMPORT :: name1, name2"),
                          ("IMPORT::name1,name2", "IMPORT :: name1, name2"),
                          ("  IMPORT  ::  name1  ,  name2  ",
                           "IMPORT :: name1, name2")])
def test_match_valid(example, result):
    ''' Test that valid input is parsed correctly '''

    obj = Import_Stmt(example)
    assert isinstance(obj, Import_Stmt), repr(obj)
    assert str(obj) == result


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize("example", ["", "  ", "IMPOR", "IMPORT : name1",
                                     "IMPORT ::", "IMPORTname1"])
def test_match_invalid(example):
    ''' Test that invalid input raises an exception '''

    with pytest.raises(NoMatchError) as excinfo:
        _ = Import_Stmt(example)
    assert "Import_Stmt: '{0}'".format(example) in \
        str(excinfo.value)



# Copyright (c) 2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Module containing pytest tests for the fparser2 Base class. '''

import pytest
from fparser.api import get_reader
from fparser.two import Fortran2003
from fparser.two.utils import walk


TEST_CODE = ("program hello\n"
             "  implicit none\n"
             "  integer :: var1, ji\n"
             "  real(wp), dimension(10,10) :: var2\n"
             "  write(*,*) 'Guten Tag'\n"
             "  do ji = 1, var1\n"
             "    var2(ji, 5) = -1.0\n"
             "  end do\n"
             "  if(var1 < 3)then\n"
             "    call a_routine(var2)\n"
             "  end if\n"
             "end program hello\n")


@pytest.mark.usefixtures("f2003_create")
def test_parent_info():
    ''' Check that parent information is correctly set-up in the
    parse tree. '''
    from fparser.two.utils import Base
    reader = get_reader(TEST_CODE)
    main = Fortran2003.Program(reader)
    node_list = walk(main)

    # Root node in the parse tree has no parent
    parent_prog = node_list[0]
    assert parent_prog.parent is None

    # Check connectivity of all non-string nodes
    for node in node_list[1:]:
        if isinstance(node, Base):
            for child in node.children:
                if isinstance(child, Base):
                    assert child.parent is node
                    assert child.get_root() is parent_prog


@pytest.mark.usefixtures("f2003_create")
def test_children_property():
    ''' Test that the children property of Base returns the correct
    results for both statements and expressions. '''
    reader = get_reader(TEST_CODE)
    main = Fortran2003.Program(reader)
    # Check that children returns items when we have an expression
    writes = walk(main, Fortran2003.Write_Stmt)
    assert writes[0].children is writes[0].items
    assert len(writes[0].children) == 2
    # Check that it returns content when we have a statement
    do_stmts = walk(main, Fortran2003.Block_Nonlabel_Do_Construct)
    assert do_stmts[0].children is do_stmts[0].content
    assert len(do_stmts[0].children) == 3



# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Fortran single line statements.

"""

from __future__ import print_function

import re
import sys

from fparser.common.base_classes import Statement

# Auxiliary tools

from fparser.common.utils import split_comma, specs_split_comma, \
                                 AnalyzeError, ParseError,       \
                                 parse_bind, parse_result,       \
                                 is_name, extract_bracketed_list_items
from fparser.common.utils import classes

__all__ = ['GeneralAssignment',
           'Assignment', 'PointerAssignment', 'Assign', 'Call', 'Goto',
           'ComputedGoto', 'AssignedGoto', 'Continue', 'Return', 'Stop',
           'Print', 'Read', 'Read0', 'Read1', 'Write', 'Flush', 'Wait',
           'Contains', 'Allocate', 'Deallocate', 'ModuleProcedure', 'Access',
           'Public', 'Private', 'Close', 'Cycle', 'Backspace', 'Endfile',
           'Rewind', 'Open', 'Format', 'Save', 'Data', 'Nullify', 'Use',
           'Exit', 'Parameter', 'Equivalence', 'Dimension', 'Target',
           'Pointer', 'Protected', 'Volatile', 'Value', 'ArithmeticIf',
           'Intrinsic', 'Inquire', 'Sequence', 'External', 'Namelist',
           'Common', 'Optional', 'Intent', 'Entry', 'Import', 'ForallStmt',
           'SpecificBinding', 'GenericBinding', 'FinalBinding', 'Allocatable',
           'Asynchronous', 'Bind', 'Else', 'ElseIf', 'Case', 'TypeIs',
           'ClassIs', 'WhereStmt', 'ElseWhere', 'Enumerator', 'FortranName',
           'Threadsafe', 'Depend', 'Check', 'CallStatement',
           'CallProtoArgument', 'Pause', 'Comment']


class StatementWithNamelist(Statement):
    """
    <statement> [ :: ] <name-list>
    """
    def process_item(self):
        if self.item.has_map():
            self.isvalid = False
            return
        if hasattr(self, 'stmtname'):
            clsname = self.stmtname
        else:
            clsname = self.__class__.__name__
        line = self.item.get_line()[len(clsname):].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = items = []
        for item in split_comma(line):
            if not is_name(item):
                self.isvalid = False
                return
            items.append(item)
        return

    def tofortran(self, isfix=None):
        if hasattr(self, 'stmtname'):
            clsname = self.stmtname.upper()
        else:
            clsname = self.__class__.__name__.upper()
        s = ', '.join(self.items)
        if s:
            s = ' ' + s
        return self.get_indent_tab(isfix=isfix) + clsname + s

# Execution statements


class GeneralAssignment(Statement):
    """
    <variable> = <expr>
    <pointer variable> => <expr>
    """

    match = re.compile(r'\w[^=]*\s*=\>?').match
    item_re = re.compile(
        r'(?P<variable>\w[^=]*)\s*(?P<sign>=\>?)\s*(?P<expr>.*)\Z', re.I).match
    _repr_attr_names = ['variable', 'sign', 'expr'] \
        + Statement._repr_attr_names

    def process_item(self):
        m = self.item_re(self.item.get_line())
        if not m:
            self.isvalid = False
            return
        self.sign = sign = m.group('sign')
        if isinstance(self, Assignment) and sign != '=':
            self.isvalid = False
            return
        elif isinstance(self, PointerAssignment) and sign != '=>':
            self.isvalid = False
            return
        else:
            if sign == '=>':
                self.__class__ = PointerAssignment
            else:
                self.__class__ = Assignment
        apply_map = self.item.apply_map
        v1 = v = m.group('variable').replace(' ', '')
        while True:
            i = v.find(')')
            if i == -1:
                break
            v = v[i+1:]
            if v.startswith('(') or v.startswith(r'%'):
                continue
            if v:
                self.isvalid = False
                return
        self.variable = apply_map(v1)
        self.expr = apply_map(m.group('expr'))
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + '%s %s %s' \
               % (self.variable, self.sign, self.expr)

    def analyze(self):
        return


class Assignment(GeneralAssignment):
    pass


class PointerAssignment(GeneralAssignment):
    pass


class Assign(Statement):
    """
    ASSIGN <label> TO <int-variable-name>
    """
    modes = ['fix77']
    match = re.compile(r'assign\s*\d+\s*to\s*\w+\s*\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[6:].lstrip()
        i = line.lower().find('to')
        assert not self.item.has_map()
        self.items = [line[:i].rstrip(), line[i+2:].lstrip()]
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'ASSIGN %s TO %s' \
               % (self.items[0], self.items[1])

    def analyze(self):
        return


class Call(Statement):
    """Call statement class
    CALL <procedure-designator> [ ( [ <actual-arg-spec-list> ] ) ]

    <procedure-designator> = <procedure-name>
                           | <proc-component-ref>
                           | <data-ref> % <binding-name>

    <actual-arg-spec> = [ <keyword> = ] <actual-arg>
    <actual-arg> = <expr>
                 | <variable>
                 | <procedure-name>
                 | <proc-component-ref>
                 | <alt-return-spec>
    <alt-return-spec> = * <label>

    <proc-component-ref> = <variable> % <procedure-component-name>

    <variable> = <designator>

    Call instance has attributes:
      designator
      arg_list
    """
    # As indicated in the specification above, a call to a subroutine
    # that has no arguments does *not* require parentheses.
    # e.g.:
    #   call bob
    # is valid Fortran.
    match = re.compile(r'call\b\s*\w([\s\w\(\)\%]*\w)?\s*(\(.*\))?\s*$',
                       re.I).match

    def process_item(self):
        ''' Parse the string containing the Call and store the
        designator and list of arguments (if any) '''
        item = self.item
        apply_map = item.apply_map
        line = item.get_line()[4:].strip()
        # Work backwards from the end of the line in order to allow
        # for code like:
        #     call my_type(1)%my_function(arg(2))
        # The following code will also support something like:
        #     call my_type(1)%my_function("(")
        # because fparser will previously have identified the "(" as a
        # string expression and replaced it with something like
        # "F2PY_EXPR_TUPLE_2"
        if line.endswith(')'):
            # Work back down the line until we find the matching '('
            i = len(line) - 2
            nopen = 1
            while i > 0:
                if line[i] == ')':
                    nopen += 1
                elif line[i] == '(':
                    nopen -= 1
                if nopen == 0:
                    # Have found the matching '(' at position i
                    break
                i -= 1
            if i <= 0:
                # Have reached the beginning of the string without
                # finding the matching '('
                self.isvalid = False
                return
            self.designator = apply_map(line[:i]).strip()
            items = split_comma(line[i+1:-1], item)
        else:
            # Call has no argument list
            items = []
            self.designator = apply_map(line).strip()
        self.items = items
        return

    def tofortran(self, isfix=None):
        ''' Returns the Fortran representation of this object as a string '''
        txt = self.get_indent_tab(isfix=isfix) + 'CALL ' + str(self.designator)
        if self.items:
            txt += '(' + ', '.join(map(str, self.items)) + ')'
        return txt

    def analyze(self):
        a = self.programblock.a
        variables = a.variables
        if hasattr(a, 'external'):
            external = a.external
            if self.designator in external:
                print('Need to analyze:', self, file=sys.stderr)
        return


class Goto(Statement):
    """
    GO TO <label>
    """
    match = re.compile(r'go\s*to\s*\d+\s*\Z', re.I).match

    def process_item(self):
        assert not self.item.has_map()
        self.label = self.item.get_line()[2:].lstrip()[2:].lstrip()
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'GO TO %s' % (self.label)

    def analyze(self):
        return


class ComputedGoto(Statement):
    """
    GO TO ( <label-list> ) [ , ] <scalar-int-expr>
    """
    match = re.compile(r'go\s*to\s*\(', re.I).match

    def process_item(self):
        apply_map = self.item.apply_map
        line = self.item.get_line()[2:].lstrip()[2:].lstrip()
        i = line.index(')')
        self.items = split_comma(line[1:i], self.item)
        line = line[i+1:].lstrip()
        if line.startswith(','):
            line = line[1:].lstrip()
        self.expr = apply_map(line)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'GO TO (%s) %s' \
               % (', '.join(self.items), self.expr)

    def analyze(self):
        return


class AssignedGoto(Statement):
    """
    GO TO <int-variable-name> [ ( <label> [ , <label> ]... ) ]
    """
    modes = ['fix77']
    match = re.compile(r'go\s*to\s*\w+\s*\(?', re.I).match

    def process_item(self):
        line = self.item.get_line()[2:].lstrip()[2:].lstrip()
        i = line.find('(')
        if i == -1:
            self.varname = line
            self.items = []
            return
        self.varname = line[:i].rstrip()
        assert line[-1] == ')', repr(line)
        self
        self.items = split_comma(line[i+1:-1], self.item)
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        if self.items:
            return tab + 'GO TO %s (%s)' \
                   % (self.varname, ', '.join(self.items))
        return tab + 'GO TO %s' % (self.varname)

    def analyze(self): return


class Continue(Statement):
    """
    CONTINUE
    """
    match = re.compile(r'continue\Z', re.I).match

    def process_item(self):
        self.label = self.item.label
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(deindent=True) + 'CONTINUE'

    def analyze(self):
        return


class Return(Statement):
    """
    RETURN [ <scalar-int-expr> ]
    """
    match = re.compile(r'return\b', re.I).match

    def process_item(self):
        self.expr = self.item.apply_map(self.item.get_line()[6:].lstrip())
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        if self.expr:
            return tab + 'RETURN %s' % (self.expr)
        return tab + 'RETURN'

    def analyze(self):
        return


class Stop(Statement):
    """
    STOP [ <stop-code> ]
    <stop-code> = <scalar-char-constant> | <1-5-digit>
    """
    match = re.compile(r'stop\s*((\'\w*\'|"\w*")+|\d+|)\Z', re.I).match

    def process_item(self):
        self.code = self.item.apply_map(self.item.get_line()[4:].lstrip())
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        if self.code:
            return tab + 'STOP %s' % (self.code)
        return tab + 'STOP'

    def analyze(self):
        return


class Print(Statement):
    """
    PRINT <format> [, <output-item-list>]
    <format> = <default-char-expr> | <label> | *

    <output-item> = <expr> | <io-implied-do>
    <io-implied-do> = ( <io-implied-do-object-list> , <implied-do-control> )
    <io-implied-do-object> = <input-item> | <output-item>
    <implied-do-control> = <do-variable>
                         = <scalar-int-expr> ,
                           <scalar-int-expr> [ , <scalar-int-expr> ]
    <input-item> = <variable> | <io-implied-do>
    """
    match = re.compile(r'print\s*(\'\w*\'|\"\w*\"|\d+|[*]|\b\w)', re.I).match

    def process_item(self):
        item = self.item
        apply_map = item.apply_map
        line = item.get_line()[5:].lstrip()
        items = split_comma(line, item)
        self.format = items[0]
        self.items = items[1:]
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'PRINT %s' \
               % (', '.join([self.format]+self.items))

    def analyze(self):
        return


class Read(Statement):
    """
Read0:    READ ( <io-control-spec-list> ) [ <input-item-list> ]

    <io-control-spec-list> = [ UNIT = ] <io-unit>
                             | [ FORMAT = ] <format>
                             | [ NML = ] <namelist-group-name>
                             | ADVANCE = <scalar-default-char-expr>
                             ...

Read1:    READ <format> [, <input-item-list>]
    <format> == <default-char-expr> | <label> | *
    """
    match = re.compile(r'read\b\s*[\w(*\'"]', re.I).match

    def process_item(self):
        item = self.item
        line = item.get_line()[4:].lstrip()
        if line.startswith('('):
            self.__class__ = Read0
        else:
            self.__class__ = Read1
        self.process_item()
        return

    def analyze(self):
        return


class Read0(Read):

    def process_item(self):
        item = self.item
        line = item.get_line()[4:].lstrip()
        i = line.find(')')
        self.specs = specs_split_comma(line[1:i], item)
        self.items = split_comma(line[i+1:], item)
        return

    def tofortran(self, isfix=None):
        s = self.get_indent_tab(isfix=isfix) \
            + 'READ (%s)' % (', '.join(self.specs))
        if self.items:
            return s + ' ' + ', '.join(self.items)
        return s


class Read1(Read):

    def process_item(self):
        item = self.item
        line = item.get_line()[4:].lstrip()
        items = split_comma(line, item)
        self.format = items[0]
        self.items = items[1:]
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'READ ' \
               + ', '.join([self.format]+self.items)


class Write(Statement):
    """
    WRITE ( io-control-spec-list ) [<output-item-list>]
    """
    match = re.compile(r'write\s*\(', re.I).match

    def process_item(self):
        item = self.item
        line = item.get_line()[5:].lstrip()
        i = line.find(')')
        assert i != -1, repr(line)
        self.specs = specs_split_comma(line[1:i], item)
        self.items = split_comma(line[i+1:], item)
        return

    def tofortran(self, isfix=None):
        s = self.get_indent_tab(isfix=isfix) \
            + 'WRITE (%s)' % ', '.join(self.specs)
        if self.items:
            s += ' ' + ', '.join(self.items)
        return s

    def analyze(self):
        return


class Flush(Statement):
    """
    FLUSH <file-unit-number>
    FLUSH ( <flush-spec-list> )
    <flush-spec> = [ UNIT = ] <file-unit-number>
                 | IOSTAT = <scalar-int-variable>
                 | IOMSG = <iomsg-variable>
                 | ERR = <label>
    """
    match = re.compile(r'flush\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[5:].lstrip()
        if not line:
            self.isvalid = False
            return
        if line.startswith('('):
            assert line[-1] == ')', repr(line)
            self.specs = specs_split_comma(line[1:-1], self.item)
        else:
            self.specs = specs_split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'FLUSH (%s)' % (', '.join(self.specs))

    def analyze(self):
        return


class Wait(Statement):
    """
    WAIT ( <wait-spec-list> )
    <wait-spec> = [ UNIT = ] <file-unit-number>
                | END = <label>
                | EOR = <label>
                | ERR = <label>
                | ID = <scalar-int-expr>
                | IOMSG = <iomsg-variable>
                | IOSTAT = <scalar-int-variable>

    """
    match = re.compile(r'wait\s*\(.*\)\Z', re.I).match

    def process_item(self):
        self.specs = specs_split_comma(
            self.item.get_line()[4:].lstrip()[1:-1], self.item)
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'WAIT (%s)' % (', '.join(self.specs))

    def analyze(self):
        return


class Contains(Statement):
    """
    CONTAINS
    """
    match = re.compile(r'contains\Z', re.I).match

    def process_item(self):
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'CONTAINS'


class Allocate(Statement):
    """
    ALLOCATE ( [ <type-spec> :: ] <allocation-list> [ , <alloc-opt-list> ] )
    <alloc-opt> = STAT = <stat-variable>
                | ERRMSG = <errmsg-variable>
                | SOURCE = <source-expr>
    <allocation> = <allocate-object> [ ( <allocate-shape-spec-list> ) ]
    """
    match = re.compile(r'allocate\s*\(.*\)\Z', re.I).match

    def process_item(self):
        '''
        Process the ALLOCATE statement and store the various entities being
        allocated in self.items. Any type-specification is stored in
        self.spec.

        :raises ParseError: if an invalid type-specification is used
        '''
        line = self.item.get_line()[8:].lstrip()[1:-1].strip()
        item2 = self.item.copy(line, True)
        line2 = item2.get_line()
        i = line2.find('::')
        if i != -1:
            spec = item2.apply_map(line2[:i].rstrip())
            from .block_statements import type_spec
            stmt = None
            for cls in type_spec:
                if cls.match(spec):
                    stmt = cls(self, item2.copy(spec))
                    if stmt.isvalid:
                        break
            if stmt is not None and stmt.isvalid:
                spec = stmt
            elif is_name(spec):
                # Type spec is the name of a derived type
                pass
            else:
                raise ParseError(
                    "Unrecognised type-specification in ALLOCATE statement: "
                    "{0}".format(self.item.line))
            line2 = line2[i+2:].lstrip()
        else:
            spec = None
        self.spec = spec
        self.items = specs_split_comma(line2, item2)
        return

    def tofortran(self, isfix=None):
        '''
        Create the Fortran code for this ALLOCATE statement

        :param bool isfix: whether or not to generate fixed-format code
        :return: Fortran code
        :rtype: str
        '''
        import six
        type_spec = ''
        if self.spec:
            if isinstance(self.spec, six.text_type):
                type_spec = self.spec
            else:
                type_spec = self.spec.tostr()
            type_spec += ' :: '
        return self.get_indent_tab(isfix=isfix) \
            + 'ALLOCATE (%s%s)' % (type_spec, ', '.join(self.items))

    def analyze(self):
        return


class Deallocate(Statement):
    """
    DEALLOCATE ( <allocate-object-list> [ , <dealloc-opt-list> ] )
    <allocate-object> = <variable-name>
                      | <structure-component>
    <structure-component> = <data-ref>
    <dealloc-opt> = STAT = <stat-variable>
                    | ERRMSG = <errmsg-variable>
    """
    match = re.compile(r'deallocate\s*\(.*\)\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[10:].lstrip()[1:-1].strip()
        self.items = specs_split_comma(line, self.item)
        return

    def tofortran(self, isfix=None): return self.get_indent_tab(isfix=isfix) \
        + 'DEALLOCATE (%s)' % (', '.join(self.items))

    def analyze(self):
        return


class ModuleProcedure(Statement):
    """
    [ MODULE ] PROCEDURE [::] <procedure-name-list>
    """
    match = re.compile(r'(module\s*|)procedure\b\s*(::)?', re.I).match

    def process_item(self):
        line = self.item.get_line()
        m = self.match(line)
        assert m, repr(line)
        items = split_comma(line[m.end():].strip(), self.item)
        for n in items:
            if not is_name(n):
                self.isvalid = False
                return
        self.items = items
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'MODULE PROCEDURE %s' % (', '.join(self.items))

    def analyze(self):
        module_procedures = self.parent.a.module_procedures
        module_procedures.extend(self.items)
        # XXX: add names to parent_provides
        return


class Access(Statement):
    """
    <access-spec> [ [::] <access-id-list>]
    <access-spec> = PUBLIC | PRIVATE
    <access-id> = <use-name> | <generic-spec>
    """
    match = re.compile(r'(public|private)\b', re.I).match

    def process_item(self):
        clsname = self.__class__.__name__.lower()
        line = self.item.get_line()
        if not line.lower().startswith(clsname):
            self.isvalid = False
            return
        line = line[len(clsname):].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        clsname = self.__class__.__name__.upper()
        tab = self.get_indent_tab(isfix=isfix)
        if self.items:
            return tab + clsname + ' ' + ', '.join(self.items)
        return tab + clsname

    def analyze(self):
        clsname = self.__class__.__name__
        bits = getattr(self.parent.a, clsname.lower() + '_id_list')
        if self.items:
            for name in self.items:
                if name not in bits:
                    bits.append(name)
        else:
            if '' not in bits:
                bits.append('')
            if not isinstance(self.parent, classes.Module):
                parentclsname = self.parent.__class__.__name__
                message = 'C548 violation: %s statement only allowed in the'\
                          ' specification-part of a module, not in a %s.'\
                          % (clsname.upper(), parentclsname.lower())
                self.warning(message)
        access_id_list = self.parent.a.private_id_list \
            + self.parent.a.public_id_list
        if access_id_list.count('') > 1:
            message = 'C548 violation: only one access-stmt with omitted'\
                      ' access-id-list is permitted in'\
                      ' the module-specification-part.'
            self.warning(message)
        # todo: check for conflicting access statement usages
        #       (e.g. private foo; public foo)
        # todo: check for conflicting generic-spec id-s.
        return


class Public(Access):
    is_public = True


class Private(Access):
    is_public = False


class Close(Statement):
    """
    CLOSE ( <close-spec-list> )
    <close-spec> = [ UNIT = ] <file-unit-number>
                   | IOSTAT = <scalar-int-variable>
                   | IOMSG = <iomsg-variable>
                   | ERR = <label>
                   | STATUS = <scalar-default-char-expr>
    """
    match = re.compile(r'close\s*\(.*\)\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[5:].lstrip()[1:-1].strip()
        self.specs = specs_split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'CLOSE (%s)' % (', '.join(self.specs))

    def analyze(self):
        return


class Cycle(Statement):
    """
    CYCLE [ <do-construct-name> ]
    """
    match = re.compile(r'cycle\b\s*\w*\s*\Z', re.I).match

    def process_item(self):
        self.name = self.item.get_line()[5:].lstrip()
        return

    def tofortran(self, isfix=None):
        if self.name:
            return self.get_indent_tab(isfix=isfix) + 'CYCLE ' + self.name
        return self.get_indent_tab(isfix=isfix) + 'CYCLE'

    def analyze(self):
        return


class FilePositioningStatement(Statement):
    """
    REWIND <file-unit-number>
    REWIND ( <position-spec-list> )
    <position-spec-list> = [ UNIT = ] <file-unit-number>
                           | IOMSG = <iomsg-variable>
                           | IOSTAT = <scalar-int-variable>
                           | ERR = <label>
    The same for BACKSPACE, ENDFILE.
    """
    match = re.compile(r'(rewind|backspace|endfile)\b', re.I).match

    def process_item(self):
        clsname = self.__class__.__name__.lower()
        line = self.item.get_line()
        if not line.lower().startswith(clsname):
            self.isvalid = False
            return
        line = line[len(clsname):].lstrip()
        if line.startswith('('):
            assert line[-1] == ')', repr(line)
            spec = line[1:-1].strip()
        else:
            spec = line
        self.specs = specs_split_comma(spec, self.item)
        return

    def tofortran(self, isfix=None):
        clsname = self.__class__.__name__.upper()
        return self.get_indent_tab(isfix=isfix) + clsname \
            + ' (%s)' % (', '.join(self.specs))

    def analyze(self):
        return


class Backspace(FilePositioningStatement):
    pass


class Endfile(FilePositioningStatement):
    pass


class Rewind(FilePositioningStatement):
    pass


class Open(Statement):
    """
    OPEN ( <connect-spec-list> )
    <connect-spec> = [ UNIT = ] <file-unit-number>
                     | ACCESS = <scalar-default-char-expr>
                     | ..
    """
    match = re.compile(r'open\s*\(.*\)\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[4:].lstrip()[1:-1].strip()
        self.specs = specs_split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'OPEN (%s)' % (', '.join(self.specs))

    def analyze(self):
        return


class Format(Statement):
    """
    FORMAT <format-specification>
    <format-specification> = ( [ <format-item-list> ] )
    <format-item> = [ <r> ] <data-edit-descr>
                    | <control-edit-descr>
                    | <char-string-edit-descr>
                    | [ <r> ] ( <format-item-list> )
    <data-edit-descr> = I <w> [ . <m> ]
                        | B <w> [ . <m> ]
                        ...
    <r|w|m|d|e> = <int-literal-constant>
    <v> = <signed-int-literal-constant>
    <control-edit-descr> = <position-edit-descr>
                         | [ <r> ] /
                         | :
                         ...
    <position-edit-descr> = T <n>
                            | TL <n>
                            ...
    <sign-edit-descr> = SS | SP | S
    ...

    """
    match = re.compile(r'format\s*\(.*\)\Z', re.I).match

    def process_item(self):
        item = self.item
        if item.label is None:
            # R1001:
            self.warning('FORMAT statement must be labeled (F2008:C1001).'
                         % (item.label))
        line = item.get_line()[6:].lstrip()
        assert line[0]+line[-1] == '()', repr(line)
        self.specs = split_comma(line[1:-1], item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'FORMAT (%s)' % (', '.join(self.specs))

    def analyze(self):
        return


class Save(Statement):
    """
    SAVE [ [ :: ] <saved-entity-list> ]
    <saved-entity> = <object-name>
                     | <proc-pointer-name>
                     | / <common-block-name> /
    <proc-pointer-name> = <name>
    <object-name> = <name>
    """
    match = re.compile(r'save\b', re.I).match

    def process_item(self):
        assert not self.item.has_map()
        line = self.item.get_line()[4:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        items = []
        for s in line.split(','):
            s = s.strip()
            if not s:
                continue
            if s.startswith('/'):
                assert s.endswith('/'), repr(s)
                n = s[1:-1].strip()
                assert is_name(n), repr(n)
                items.append('/%s/' % (n))
            elif is_name(s):
                items.append(s)
            else:
                self.isvalid = False
                return
        self.items = items
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        if not self.items:
            return tab + 'SAVE'
        return tab + 'SAVE %s' % (', '.join(self.items))

    def analyze(self):
        return


class Data(Statement):
    """
    DATA <data-stmt-set> [ [ , ] <data-stmt-set> ]...
    <data-stmt-set> = <data-stmt-object-list> / <data-stmt-value-list> /
    <data-stmt-object> = <variable> | <data-implied-do>
    <data-implied-do> = ( <data-i-do-object-list> ,
                          <data-i-do-variable> = <scalar-int-expr> ,
                          <scalar-int-expr> [ , <scalar-int-expr> ] )
    <data-i-do-object> = <array-element>
                         | <scalar-structure-component>
                         | <data-implied-do>
    <data-i-do-variable> = <scalar-int-variable>
    <variable> = <designator>
    <designator> = <object-name>
                   | <array-element>
                   | <array-section>
                   | <structure-component>
                   | <substring>
    <array-element> = <data-ref>
    <array-section> = <data-ref> [ ( <substring-range> ) ]

    """
    match = re.compile(r'data\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[4:].lstrip()
        stmts = []
        self.isvalid = False
        while line:
            i = line.find('/')
            if i == -1:
                return
            j = line.find('/', i+1)
            if j == -1:
                return
            l1, l2 = line[:i].rstrip(), line[i+1:j].strip()
            l1 = split_comma(l1, self.item)
            l2 = split_comma(l2, self.item)
            stmts.append((l1, l2))
            line = line[j+1:].lstrip()
            if line.startswith(','):
                line = line[1:].lstrip()
        self.stmts = stmts
        self.isvalid = True
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        bits = []
        for o, v in self.stmts:
            bits.append('%s / %s /' % (', '.join(o), ', '.join(v)))
        return tab + 'DATA ' + ' '.join(bits)

    def analyze(self):
        return


class Nullify(Statement):
    """
    NULLIFY ( <pointer-object-list> )
    <pointer-object> = <variable-name>
    """
    match = re.compile(r'nullify\s*\(.*\)\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[7:].lstrip()[1:-1].strip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'NULLIFY (%s)' % (', '.join(self.items))

    def analyze(self):
        return


class Use(Statement):
    """Parses USE statement.

    :param class Statement: Base fparser class.
    :raises AnalyzeError: If entity name is not in module.

    Fortran syntax construct:

    USE [ [ , <module-nature> ] :: ] <module-name> [ , <rename-list> ]
    USE [ [ , <module-nature> ] :: ] <module-name> , ONLY : [ <only-list> ]
    <module-nature> = INTRINSIC | NON_INTRINSIC
    <rename> = <local-name> => <use-name>
               | OPERATOR ( <local-defined-operator> ) =>
                            OPERATOR ( <use-defined-operator> )
    <only> = <generic-spec> | <only-use-name> | <rename>
    <only-use-name> = <use-name>
    """
    match = re.compile(r'use\b', re.I).match

    def process_item(self):
        ''' Parse the string containing the Use and store the
        module name and list of attributes (if any) '''
        line = self.item.get_line()[3:].lstrip()
        nature = ''
        if line.startswith(','):
            i = line.find('::')
            nature = line[1:i].strip().upper()
            line = line[i+2:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        if nature and not is_name(nature):
            self.isvalid = False
            return
        self.nature = nature
        i = line.find(',')
        self.isonly = False
        if i == -1:
            self.name = line
            self.items = []
        else:
            self.name = line[:i].rstrip()
            line = line[i+1:].lstrip()
            if line.lower().startswith('only') and \
               line[4:].lstrip().startswith(':'):
                self.isonly = True
                line = line[4:].lstrip()[1:].lstrip()
            self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):  # pylint: disable=invalid-name
        '''
        Returns the Fortran representation of this object as a string

        :param bool isfix: Whether or not to generated fixed-format Fortran
        :return: Fortran representation of this object
        :rtype: str
        '''
        tab = self.get_indent_tab(isfix=isfix)
        s = 'USE'
        if self.nature:
            s += ', ' + self.nature + ' ::'
        s += ' ' + self.name
        if self.isonly:
            s += ', ONLY:'
        elif self.items:
            s += ','
        if self.items:
            s += ' ' + ', '.join(self.items)
        return tab + s

    def analyze(self):  # pylint: disable=invalid-name
        ''' Returns warnings if this object is incorrect '''
        use = self.parent.a.use
        if self.name in use:
            return

        modules = self.top.a.module
        if self.name not in modules:
            fn = self.reader.find_module_source_file(self.name)
            if fn is not None:
                from fparser.common.readfortran import FortranFileReader
                from fparser.one.parsefortran import FortranParser
                self.info('looking module information from %r' % (fn))
                reader = FortranFileReader(
                    fn, include_dirs=self.reader.include_dirs,
                    source_only=self.reader.source_only)
                parser = FortranParser(reader)
                parser.parse()
                parser.block.a.module.update(modules)
                parser.analyze()
                modules.update(parser.block.a.module)

        if self.name not in modules:
            self.warning('no information about the module %r in use statement'
                         % (self.name))
            return

        module = modules[self.name]
        use[self.name] = module
        use_provides = self.parent.a.use_provides
        renames = [split_comma(item, comma='=>')
                   for item in self.items if '=>' in item]
        norenames = [item for item in self.items if '=>' not in item]
        all_mod_provides = dict(module.a.module_provides)
        all_mod_provides.update(module.a.use_provides)
        if self.isonly:
            # populate use_provides with items/renames only.
            for rename, orig in renames:
                self.populate_use_provides(
                    all_mod_provides, use_provides, orig, rename)
            for name in norenames:
                self.populate_use_provides(
                    all_mod_provides, use_provides, name)
        else:
            # norenames should be empty
            if norenames:
                self.warning("'use' without 'only' clause does not rename the "
                             "variables '%s'" % ', '.join(norenames))
            # populate use_provides with renamed vars from module.
            for rename, orig in renames:
                self.populate_use_provides(
                    all_mod_provides, use_provides, orig, rename)
            # get all the rest
            unrenamed = set(all_mod_provides) - set([b for (a, b) in renames])
            for name in unrenamed:
                self.populate_use_provides(
                    all_mod_provides, use_provides, name)
        return

    def populate_use_provides(self, all_mod_provides, use_provides, name,
                              rename=None):
        ''' Checks for entity name in the module '''
        ovar = all_mod_provides.get(name, None)
        if ovar is None:
            raise AnalyzeError("entity name '%s' is not in module '%s'"
                               % (name, self.name))
        if rename:
            # XXX: rename != ovar.name -- should mark this somehow?
            name_idx = rename
        else:
            name_idx = name
        if name_idx in use_provides:
            if ovar != use_provides[name_idx]:
                self.warning("entity name '%s' is already declared in module "
                             "'%s' while adding it to '%s', overriding."
                             % (name, self.name, self.parent.name))
        use_provides[name_idx] = ovar


class Exit(Statement):
    """
    EXIT [ <do-construct-name> ]
    """
    match = re.compile(r'exit\b\s*\w*\s*\Z', re.I).match

    def process_item(self):
        self.name = self.item.get_line()[4:].lstrip()
        return

    def tofortran(self, isfix=None):
        if self.name:
            return self.get_indent_tab(isfix=isfix) + 'EXIT ' + self.name
        return self.get_indent_tab(isfix=isfix) + 'EXIT'

    def analyze(self):
        return


class Parameter(Statement):
    """
    PARAMETER ( <named-constant-def-list> )
    <named-constant-def> = <named-constant> = <initialization-expr>
    """
    match = re.compile(r'parameter\s*\(.*\)\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[9:].lstrip()[1:-1].strip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'PARAMETER (%s)' % (', '.join(self.items))

    def analyze(self):
        for item in self.items:
            i = item.find('=')
            assert i != -1, repr(item)
            name = item[:i].rstrip()
            value = item[i+1:].lstrip()
            var = self.get_variable(name)
            var.update('parameter')
            var.set_init(value)
        return


class Equivalence(Statement):
    """
    EQUIVALENCE <equivalence-set-list>
    <equivalence-set> = ( <equivalence-object> , <equivalence-object-list> )
    <equivalence-object> = <variable-name> | <array-element> | <substring>
    """
    match = re.compile(r'equivalence\s*\(.*\)\Z', re.I).match

    def process_item(self):
        items = []
        for s in self.item.get_line()[11:].lstrip().split(','):
            s = s.strip()
            assert s[0]+s[-1] == '()', repr((s, self.item.get_line()))
            s = ', '.join(split_comma(s[1:-1], self.item))
            items.append('('+s+')')
        self.items = items
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'EQUIVALENCE %s' % (', '.join(self.items))

    def analyze(self):
        return


class Dimension(Statement):
    """
    DIMENSION [ :: ] <array-name> ( <array-spec> )
                                    [ , <array-name> ( <array-spec> ) ]...

    """
    match = re.compile(r'dimension\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[9:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'DIMENSION %s' % (', '.join(self.items))

    def analyze(self):
        for line in self.items:
            i = line.find('(')
            assert i != -1 and line.endswith(')'), repr(line)
            name = line[:i].rstrip()
            array_spec = split_comma(line[i+1:-1].strip(), self.item)
            var = self.get_variable(name)
            var.set_bounds(array_spec)
        return


class Target(Statement):
    """
    TARGET [ :: ] <object-name> ( <array-spec> )
                                  [ , <object-name> ( <array-spec> ) ]...

    """
    match = re.compile(r'target\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[6:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'TARGET %s' % (', '.join(self.items))

    def analyze(self):
        for line in self.items:
            i = line.find('(')
            assert i != -1 and line.endswith(')'), repr(line)
            name = line[:i].rstrip()
            array_spec = split_comma(line[i+1:-1].strip(), self.item)
            var = self.get_variable(name)
            var.set_bounds(array_spec)
            var.update('target')
        return


class Pointer(Statement):
    """
    POINTER [ :: ] <pointer-decl-list>
    <pointer-decl> = <object-name> [ ( <deferred-shape-spec-list> ) ]
                   | <proc-entity-name>

    """
    match = re.compile(r'pointer\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[7:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'POINTER %s' % (', '.join(self.items))

    def analyze(self):
        for line in self.items:
            i = line.find('(')
            if i == -1:
                name = line
                array_spec = None
            else:
                assert line.endswith(')'), repr(line)
                name = line[:i].rstrip()
                array_spec = split_comma(line[i+1:-1].strip(), self.item)
            var = self.get_variable(name)
            var.set_bounds(array_spec)
            var.update('pointer')
        return


class Protected(StatementWithNamelist):
    """
    PROTECTED [ :: ] <entity-name-list>
    """
    match = re.compile(r'protected\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('protected')
        return


class Volatile(StatementWithNamelist):
    """
    VOLATILE [ :: ] <object-name-list>
    """
    match = re.compile(r'volatile\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('volatile')
        return


class Value(StatementWithNamelist):
    """
    VALUE [ :: ] <dummy-arg-name-list>
    """
    match = re.compile(r'value\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('value')
        return


class ArithmeticIf(Statement):
    """
    IF ( <scalar-numeric-expr> ) <label> , <label> , <label>
    """
    match = re.compile(r'if\s*\(.*\)\s*\d+\s*,\s*\d+\s*,\s*\d+\s*\Z',
                       re.I).match

    def process_item(self):
        line = self.item.get_line()[2:].lstrip()
        line, l2, l3 = line.rsplit(',', 2)
        i = line.rindex(')')
        l1 = line[i+1:]
        self.expr = self.item.apply_map(line[1:i]).strip()
        self.labels = [l1.strip(), l2.strip(), l3.strip()]
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'IF (%s) %s' \
               % (self.expr, ', '.join(self.labels))

    def analyze(self):
        return


class Intrinsic(StatementWithNamelist):
    """
    INTRINSIC [ :: ] <intrinsic-procedure-name-list>
    """
    match = re.compile(r'intrinsic\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('intrinsic')
        return


class Inquire(Statement):
    """
    INQUIRE ( <inquire-spec-list> )
    INQUIRE ( IOLENGTH = <scalar-int-variable> ) <output-item-list>

    <inquire-spec> = [ UNIT = ] <file-unit-number>
                     | FILE = <file-name-expr>
                     ...
    <output-item> = <expr>
                  | <io-implied-do>
    """
    match = re.compile(r'inquire\s*\(', re.I).match

    def process_item(self):
        line = self.item.get_line()[7:].lstrip()
        i = line.index(')')
        self.specs = specs_split_comma(line[1:i].strip(), self.item)
        self.items = split_comma(line[i+1:].lstrip(), self.item)
        return

    def tofortran(self, isfix=None):
        if self.items:
            return self.get_indent_tab(isfix=isfix) + 'INQUIRE (%s) %s' \
                   % (', '.join(self.specs), ', '.join(self.items))
        return self.get_indent_tab(isfix=isfix) + 'INQUIRE (%s)' \
            % (', '.join(self.specs))

    def analyze(self):
        return


class Sequence(Statement):
    """
    SEQUENCE
    """
    match = re.compile(r'sequence\Z', re.I).match

    def process_item(self):
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'SEQUENCE'

    def analyze(self):
        self.parent.update_attributes('SEQUENCE')
        return


class External(StatementWithNamelist):
    """
    EXTERNAL [ :: ] <external-name-list>
    """
    match = re.compile(r'external\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('external')
        return


class Namelist(Statement):
    """
    NAMELIST / <namelist-group-name> / <namelist-group-object-list> [ [ , ]
             / <namelist-group-name> / <namelist-group-object-list> ]...
    <namelist-group-object> = <variable-name>
    """
    match = re.compile(r'namelist\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[8:].lstrip()
        items = []
        while line:
            assert line.startswith('/'), repr(line)
            i = line.find('/', 1)
            assert i != -1, repr(line)
            name = line[:i+1]
            line = line[i+1:].lstrip()
            i = line.find('/')
            if i == -1:
                items.append((name, line))
                line = ''
                continue
            s = line[:i].rstrip()
            if s.endswith(','):
                s = s[:-1].rstrip()
            items.append((name, s))
            line = line[i+1:].lstrip()
        self.items = items
        return

    def tofortran(self, isfix=None):
        bits = []
        for name, s in self.items:
            bits.append('%s %s' % (name, s))
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'NAMELIST ' + ', '.join(bits)


class Common(Statement):
    """
    COMMON [ / [ <common-block-name> ] / ]  <common-block-object-list> \
      [ [ , ] / [ <common-block-name> ] /  <common-block-object-list> ]...
    <common-block-object> = <variable-name> [ ( <explicit-shape-spec-list> ) ]
                          | <proc-pointer-name>
    """
    match = re.compile(r'common\b', re.I).match

    def process_item(self):
        item = self.item
        line = item.get_line()[6:].lstrip()
        items = []
        while line:
            if not line.startswith('/'):
                name = ''
                assert not items, repr(line)
            else:
                i = line.find('/', 1)
                assert i != -1, repr(line)
                name = line[1:i].strip()
                line = line[i+1:].lstrip()
            i = line.find('/')
            if i == -1:
                items.append((name, split_comma(line, item)))
                line = ''
                continue
            s = line[:i].rstrip()
            if s.endswith(','):
                s = s[:-1].rstrip()
            items.append((name, split_comma(s, item)))
            line = line[i:].lstrip()
        self.items = items
        return

    def tofortran(self, isfix=None):
        bits = []
        for name, s in self.items:
            s = ', '.join(s)
            if name:
                bits.append('/ %s / %s' % (name, s))
            else:
                bits.append(s)
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'COMMON ' + ' '.join(bits)

    def analyze(self):
        for cname, items in self.items:
            for item in items:
                i = item.find('(')
                if i != -1:
                    assert item.endswith(')'), repr(item)
                    name = item[:i].rstrip()
                    shape = split_comma(item[i+1:-1].strip(), self.item)
                else:
                    name = item
                    shape = None
                var = self.get_variable(name)
                if shape is not None:
                    var.set_bounds(shape)
            # XXX: add name,var to parent_provides
        return


class Optional(StatementWithNamelist):
    """
    OPTIONAL [ :: ] <dummy-arg-name-list>
    <dummy-arg-name> = <name>
    """
    match = re.compile(r'optional\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('optional')
        return


class Intent(Statement):
    """
    INTENT ( <intent-spec> ) [ :: ] <dummy-arg-name-list>
    <intent-spec> = IN | OUT | INOUT

    generalization for pyf-files:
    INTENT ( <intent-spec-list> ) [ :: ] <dummy-arg-name-list>
    <intent-spec> = IN | OUT | INOUT | CACHE | HIDE | OUT = <name>
    """
    match = re.compile(r'intent\s*\(', re.I).match

    def process_item(self):
        line = self.item.get_line()[6:].lstrip()
        i = line.find(')')
        self.specs = specs_split_comma(line[1:i], self.item, upper=True)
        line = line[i+1:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = [s.strip() for s in line.split(',')]
        for n in self.items:
            if not is_name(n):
                self.isvalid = False
                return
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'INTENT (%s) %s' \
               % (', '.join(self.specs), ', '.join(self.items))

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.set_intent(self.specs)
        return


class Entry(Statement):
    """
    ENTRY <entry-name> [ ( [ <dummy-arg-list> ] ) [ <suffix> ] ]
    <suffix> = <proc-language-binding-spec> [ RESULT ( <result-name> ) ]
             | RESULT ( <result-name> ) [ <proc-language-binding-spec> ]
    <proc-language-binding-spec> = <language-binding-spec>
    <language-binding-spec> =
                      BIND ( C [ , NAME = <scalar-char-initialization-expr> ] )
    <dummy-arg> = <dummy-arg-name> | *
    """
    match = re.compile(r'entry\s+[a-zA-Z]', re.I).match

    def process_item(self):
        line = self.item.get_line()[5:].lstrip()
        m = re.match(r'\w+', line)
        name = line[:m.end()]
        line = line[m.end():].lstrip()
        if line.startswith('('):
            i = line.find(')')
            assert i != -1, repr(line)
            items = split_comma(line[1:i], self.item)
            line = line[i+1:].lstrip()
        else:
            items = []
        self.bind, line = parse_bind(line, self.item)
        self.result, line = parse_result(line, self.item)
        if line:
            assert self.bind is None, repr(self.bind)
            self.bind, line = parse_bind(line, self.item)
        assert not line, repr(line)
        self.name = name
        self.items = items
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        s = tab + 'ENTRY '+self.name
        if self.items:
            s += ' (%s)' % (', '.join(self.items))
        if self.result:
            s += ' RESULT (%s)' % (self.result)
        if self.bind:
            s += ' BIND (%s)' % (', '.join(self.bind))
        return s


class Import(StatementWithNamelist):
    """
    IMPORT [ [ :: ] <import-name-list> ]
    """
    match = re.compile(r'import(\b|\Z)', re.I).match


class Forall(Statement):
    """
    FORALL <forall-header> <forall-assignment-stmt>
    <forall-header> = ( <forall-triplet-spec-list> [ , <scalar-mask-expr> ] )
    <forall-triplet-spec> =
                    <index-name> = <subscript> : <subscript> [ : <stride> ]
    <subscript|stride> = <scalar-int-expr>
    <forall-assignment-stmt> = <assignment-stmt> | <pointer-assignment-stmt>
    """
    match = re.compile(r'forall\s*\(.*\).*=', re.I).match

    def process_item(self):
        line = self.item.get_line()[6:].lstrip()
        i = line.index(')')

        line0 = line[1:i]
        line = line[i+1:].lstrip()
        stmt = GeneralAssignment(self, self.item.copy(line, True))
        if stmt.isvalid:
            self.content = [stmt]
        else:
            self.isvalid = False
            return

        specs = []
        mask = ''
        for l in split_comma(line0, self.item):
            j = l.find('=')
            if j == -1:
                assert not mask, repr((mask, l))
                mask = l
                continue
            assert j != -1, repr(l)
            index = l[:j].rstrip()
            it = self.item.copy(l[j+1:].lstrip())
            line = it.get_line()
            k = line.split(':')
            if len(k) == 3:
                s1, s2, s3 = list(map(it.apply_map,
                                      [k[0].strip(), k[1].strip(),
                                       k[2].strip()]))
            else:
                assert len(k) == 2, repr(k)
                s1, s2 = list(map(it.apply_map,
                                  [k[0].strip(), k[1].strip()]))
                s3 = '1'
            specs.append((index, s1, s2, s3))

        self.specs = specs
        self.mask = mask
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        lines = []
        for index, s1, s2, s3 in self.specs:
            s = '%s = %s : %s' % (index, s1, s2)
            if s3 != '1':
                s += ' : %s' % (s3)
            lines.append(s)
        s = ', '.join(lines)
        if self.mask:
            s += ', ' + self.mask
        return tab + 'FORALL (%s) %s' % \
            (s, str(self.content[0]).lstrip())

    def analyze(self):
        return


ForallStmt = Forall


class SpecificBinding(Statement):
    """
    PROCEDURE [ ( <interface-name> ) ]  [ [ , <binding-attr-list> ] :: ]
              <binding-name> [ => <procedure-name> ]
    <binding-attr> = PASS [ ( <arg-name> ) ]
                   | NOPASS
                   | NON_OVERRIDABLE
                   | DEFERRED
                   | <access-spec>
    <access-spec> = PUBLIC | PRIVATE
    """
    match = re.compile(r'procedure\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[9:].lstrip()
        if line.startswith('('):
            i = line.index(')')
            name = line[1:i].strip()
            line = line[i+1:].lstrip()
        else:
            name = ''
        self.iname = name
        if line.startswith(','):
            line = line[1:].lstrip()
        i = line.find('::')
        if i != -1:
            attrs = split_comma(line[:i], self.item)
            line = line[i+2:].lstrip()
        else:
            attrs = []
        attrs1 = []
        for attr in attrs:
            if is_name(attr):
                attr = attr.upper()
            else:
                i = attr.find('(')
                assert i != -1 and attr.endswith(')'), repr(attr)
                attr = '%s (%s)' % (attr[:i].rstrip().upper(),
                                    attr[i+1:-1].strip())
            attrs1.append(attr)
        self.attrs = attrs1
        i = line.find('=')
        if i == -1:
            self.name = line
            self.bname = ''
        else:
            self.name = line[:i].rstrip()
            self.bname = line[i+1:].lstrip()[1:].lstrip()
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        s = 'PROCEDURE '
        if self.iname:
            s += '(' + self.iname + ') '
        if self.attrs:
            s += ', ' + ', '.join(self.attrs) + ' :: '
        if self.bname:
            s += '%s => %s' % (self.name, self.bname)
        else:
            s += self.name
        return tab + s


class GenericBinding(Statement):
    """
    GENERIC [ , <access-spec> ] :: <generic-spec> => <binding-name-list>
    """
    match = re.compile(r'generic\b.*::.*=\>.*\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[7:].lstrip()
        if line.startswith(','):
            line = line[1:].lstrip()
        i = line.index('::')
        self.aspec = line[:i].rstrip().upper()
        line = line[i+2:].lstrip()
        i = line.index('=>')
        self.spec = self.item.apply_map(line[:i].rstrip())
        self.items = split_comma(line[i+2:].lstrip())
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        s = 'GENERIC'
        if self.aspec:
            s += ', '+self.aspec
        s += ' :: ' + self.spec + ' => ' + ', '.join(self.items)
        return tab + s


class FinalBinding(StatementWithNamelist):
    """
    FINAL [ :: ] <final-subroutine-name-list>
    """
    stmtname = 'final'
    match = re.compile(r'final\b', re.I).match


class Allocatable(Statement):
    """
    ALLOCATABLE [ :: ] <object-name> [ ( <deferred-shape-spec-list> ) ]
                                     [ , <object-name>
                                         [ ( <deferred-shape-spec-list> ) ]
                                     ]...
    """
    match = re.compile(r'allocatable\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[11:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) \
            + 'ALLOCATABLE ' + ', '.join(self.items)

    def analyze(self):
        for line in self.items:
            i = line.find('(')
            if i == -1:
                name = line
                array_spec = None
            else:
                assert line.endswith(')')
                name = line[:i].rstrip()
                array_spec = split_comma(line[i+1:-1], self.item)
            var = self.get_variable(name)
            var.update('allocatable')
            if array_spec is not None:
                var.set_bounds(array_spec)
        return


class Asynchronous(StatementWithNamelist):
    """
    ASYNCHRONOUS [ :: ] <object-name-list>
    """
    match = re.compile(r'asynchronous\b', re.I).match

    def analyze(self):
        for name in self.items:
            var = self.get_variable(name)
            var.update('asynchronous')
        return


class Bind(Statement):
    """
    <language-binding-spec> [ :: ] <bind-entity-list>
    <language-binding-spec> =
                      BIND ( C [ , NAME = <scalar-char-initialization-expr> ] )
    <bind-entity> = <entity-name> | / <common-block-name> /
    """
    match = re.compile(r'bind\s*\(.*\)', re.I).match

    def process_item(self):
        line = self.item.line
        self.specs, line = parse_bind(line, self.item)
        if line.startswith('::'):
            line = line[2:].lstrip()
        items = []
        for item in split_comma(line, self.item):
            if item.startswith('/'):
                assert item.endswith('/'), repr(item)
                item = '/ ' + item[1:-1].strip() + ' /'
            items.append(item)
        self.items = items
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'BIND (%s) %s' %\
               (', '.join(self.specs), ', '.join(self.items))

# IF construct statements


class Else(Statement):
    """
    ELSE [<if-construct-name>]
    """
    match = re.compile(r'else\b\s*\w*\s*\Z', re.I).match

    def process_item(self):
        item = self.item
        self.name = item.get_line()[4:].strip()
        parent_name = getattr(self.parent, 'name', '')
        if self.name and self.name != parent_name:
            self.warning('expected if-construct-name %r but got %r, skipping.'
                         % (parent_name, self.name))
            self.isvalid = False
        return

    def tofortran(self, isfix=None):
        if self.name:
            return self.get_indent_tab(deindent=True) + 'ELSE ' + self.name
        return self.get_indent_tab(deindent=True) + 'ELSE'

    def analyze(self):
        return


class ElseIf(Statement):
    """
    ELSE IF ( <scalar-logical-expr> ) THEN [ <if-construct-name> ]
    """
    match = re.compile(r'else\s*if\s*\(.*\)\s*then\s*\w*\s*\Z', re.I).match

    def process_item(self):
        item = self.item
        line = item.get_line()[4:].lstrip()[2:].lstrip()
        i = line.find(')')
        assert line[0] == '('
        self.expr = item.apply_map(line[1:i])
        self.name = line[i+1:].lstrip()[4:].strip()
        parent_name = getattr(self.parent, 'name', '')
        if self.name and self.name != parent_name:
            self.warning('expected if-construct-name %r but got %r, skipping.'
                         % (parent_name, self.name))
            self.isvalid = False
        return

    def tofortran(self, isfix=None):
        s = ''
        if self.name:
            s = ' ' + self.name
        return self.get_indent_tab(deindent=True) + 'ELSE IF (%s) THEN%s' \
            % (self.expr, s)

    def analyze(self):
        return


# SelectCase construct statements


class Case(Statement):
    """
    CASE <case-selector> [ <case-construct-name> ]
    <case-selector> = ( <case-value-range-list> ) | DEFAULT
    <case-value-range> = <case-value>
                         | <case-value> :
                         | : <case-value>
                         | <case-value> : <case-value>
    <case-value> = <scalar-(int|char|logical)-initialization-expr>
    """
    match = re.compile(r'case\b\s*(\(.*\)|DEFAULT)\s*\w*\Z', re.I).match

    def process_item(self):
        ''' Populate the state of this item by parsing the associated line
        of code '''
        line = self.item.get_line()[4:].lstrip()
        try:
            self.items = extract_bracketed_list_items(line, self.item)
            idx = line.rfind(')')
            self.name = line[idx+1:].lstrip()
        except ParseError:
            # No list in parentheses found so we must have a 'case default'
            if not line.lower().startswith('default'):
                # We should never get to here because such a string should
                # not have generated a match
                self.warning(
                    "Internal error when parsing CASE statement: {0}".
                    format(line))
                self.isvalid = False
                return
            self.items = []
            self.name = line[7:].lstrip()
        parent_name = getattr(self.parent, 'name', '')
        if self.name and self.name != parent_name:
            self.warning("Expected case-construct-name {0} but got {1}, "
                         "skipping.".format(parent_name, self.name))
            self.isvalid = False
        return

    def tofortran(self, isfix=None):
        ''' Return the Fortran for this object as a string '''
        tab = self.get_indent_tab(isfix=isfix)
        txt = tab + 'CASE'
        if self.items:
            lst = []
            for item in self.items:
                lst.append((' : '.join(item)).strip())
            txt += ' ( %s )' % (', '.join(lst))
        else:
            txt += ' DEFAULT'
        if self.name:
            txt += ' ' + self.name
        return txt

    def analyze(self):
        return


class TypeIs(Statement):
    """
    TYPE IS <type-selector> [ <case-construct-name> ]
    <type-selector> = ( <type-value-range-list> )
    <type-value-range> = <case-value>
    <case-value> = <char>
    """
    match = re.compile(r'type\b\s*is\b\s*\(.*\)\s*\w*\Z', re.I).match

    def process_item(self):
        ''' Populate the state of this object by parsing the associated
        line of code '''
        line = self.item.get_line()
        # We have a 'type is (...)' statement. At this point
        # any expression used for the type specifier will have
        # been replaced with e.g. F2PY_EXPR_TUPLE_3 and so
        # will not itself contain any parentheses.
        self.items = extract_bracketed_list_items(line, self.item)
        # Get and store the case-construct-name (if any)
        idx2 = line.rfind(')')
        self.name = line[idx2+1:].lstrip()
        parent_name = getattr(self.parent, 'name', '')
        if self.name and self.name != parent_name:
            self.warning(
                'expected type-is-construct-name %r but got %r, skipping.'
                % (parent_name, self.name))
            self.isvalid = False
        return

    def tofortran(self, isfix=None):
        ''' Create the Fortran representation of this object and return
        it as a string '''
        tab = self.get_indent_tab(isfix=isfix)
        text = tab + 'TYPE IS'
        if self.items:
            lst = []
            for item in self.items:
                lst.append((' : '.join(item)).strip())
            text += ' ( %s )' % (', '.join(lst))
        else:
            raise ParseError("TYPE IS construct must have arguments")
        if self.name:
            text += ' ' + self.name
        return text

    def analyze(self):
        return


class ClassIs(Statement):
    """
    CLASS <class-selector>
    <class-selector> = ( IS <type-value-range-list> | DEFAULT )
    <type-value-range> = <case-value>
    <case-value> = <char>
    """
    match = re.compile(r'class\b\s*(is\b\s*\(.*\)|default)\s*\w*\Z',
                       re.I).match

    def process_item(self):
        ''' Populate the state of this object by parsing the string '''
        line = self.item.get_line()[5:].lstrip()
        try:
            self.items = extract_bracketed_list_items(line, self.item)
            # We have a 'class is ...' statement. At this point
            # any expression used for the class specifier will have
            # been replaced with e.g. F2PY_EXPR_TUPLE_3 and so
            # will not contain any parentheses.
            idx = line.rfind(')')
            self.name = line[idx+1:].lstrip()
        except ParseError:
            # We have a 'class default' statement
            if not line.lower().startswith('default'):
                # We should never get here because such a string should
                # not have generated a match
                self.warning(
                    "Internal error when parsing CLASS statement: {0}".
                    format(line))
                self.isvalid = False
                return
            self.items = []
            self.name = line[7:].lstrip()
        parent_name = getattr(self.parent, 'name', '')
        if self.name and self.name != parent_name:
            self.warning(
                'expected class-construct-name %r but got %r, skipping.'
                % (parent_name, self.name))
            self.isvalid = False
        return

    def tofortran(self, isfix=None):
        ''' Returns the Fortran for this object as a string '''
        tab = self.get_indent_tab(isfix=isfix)
        text = tab + 'CLASS'
        if self.items:
            text += ' IS'
            lchar = []
            for item in self.items:
                lchar.append((' : '.join(item)).strip())
            text += ' ( %s )' % (', '.join(lchar))
        else:
            text += ' DEFAULT'
        if self.name:
            text += ' ' + self.name
        return text

    def analyze(self):
        return


# Where construct statements


class Where(Statement):
    """
    WHERE ( <mask-expr> ) <where-assignment-stmt>
    """
    match = re.compile(r'where\s*\(.*\)\s*\w.*\Z', re.I).match

    def process_item(self):
        line = self.item.get_line()[5:].lstrip()
        i = line.index(')')
        self.expr = self.item.apply_map(line[1:i].strip())
        line = line[i+1:].lstrip()
        newitem = self.item.copy(line)
        cls = Assignment
        if cls.match(line):
            stmt = cls(self, newitem)
            if stmt.isvalid:
                self.content = [stmt]
                return
        self.isvalid = False
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        return tab + 'WHERE ( %s ) %s' % (self.expr,
                                          str(self.content[0]).lstrip())

    def analyze(self):
        return


WhereStmt = Where


class ElseWhere(Statement):
    """
    ELSE WHERE ( <mask-expr> ) [ <where-construct-name> ]
    ELSE WHERE [ <where-construct-name> ]
    """
    match = re.compile(r'else\s*where\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[4:].lstrip()[5:].lstrip()
        self.expr = None
        if line.startswith('('):
            i = line.index(')')
            assert i != -1, repr(line)
            self.expr = self.item.apply_map(line[1:i].strip())
            line = line[i+1:].lstrip()
        self.name = line
        parent_name = getattr(self.parent, 'name', '')
        if self.name and not self.name == parent_name:
            self.warning('expected where-construct-name %r but got %r, '
                         'skipping.' % (parent_name, self.name))
            self.isvalid = False
        return

    def tofortran(self, isfix=None):
        tab = self.get_indent_tab(isfix=isfix)
        s = 'ELSE WHERE'
        if self.expr is not None:
            s += ' ( %s )' % (self.expr)
        if self.name:
            s += ' ' + self.name
        return tab + s

    def analyze(self):
        return

# Enum construct statements


class Enumerator(Statement):
    """
    ENUMERATOR [ :: ] <enumerator-list>
    <enumerator> = <named-constant> [ = <scalar-int-initialization-expr> ]
    """
    match = re.compile(r'enumerator\b', re.I).match

    def process_item(self):
        line = self.item.get_line()[10:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line, self.item)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'ENUMERATOR ' \
            + ', '.join(self.items)

# F2PY specific statements


class FortranName(Statement):
    """
    FORTRANNAME <name>
    """
    match = re.compile(r'fortranname\s*\w+\Z', re.I).match

    def process_item(self):
        self.value = self.item.get_line()[11:].lstrip()
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'FORTRANNAME ' \
            + self.value


class Threadsafe(Statement):
    """
    THREADSAFE
    """
    match = re.compile(r'threadsafe\Z', re.I).match

    def process_item(self):
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'THREADSAFE'


class Depend(Statement):
    """
    DEPEND ( <name-list> ) [ :: ] <dummy-arg-name-list>

    """
    match = re.compile(r'depend\s*\(', re.I).match

    def process_item(self):
        line = self.item.get_line()[6:].lstrip()
        i = line.find(')')
        self.depends = split_comma(line[1:i].strip(), self.item)
        line = line[i+1:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.items = split_comma(line)
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'DEPEND ( %s ) %s' \
               % (', '.join(self.depends), ', '.join(self.items))


class Check(Statement):
    """
    CHECK ( <c-int-scalar-expr> ) [ :: ] <name>

    """
    match = re.compile(r'check\s*\(', re.I).match

    def process_item(self):
        line = self.item.get_line()[5:].lstrip()
        i = line.find(')')
        assert i != -1, repr(line)
        self.expr = self.item.apply_map(line[1:i].strip())
        line = line[i+1:].lstrip()
        if line.startswith('::'):
            line = line[2:].lstrip()
        self.value = line
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'CHECK ( %s ) %s' \
               % (self.expr, self.value)


class CallStatement(Statement):
    """
    CALLSTATEMENT <c-expr>
    """
    match = re.compile(r'callstatement\b', re.I).match

    def process_item(self):
        self.expr = self.item.apply_map(self.item.get_line()[13:].lstrip())
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'CALLSTATEMENT ' + self.expr


class CallProtoArgument(Statement):
    """
    CALLPROTOARGUMENT <c-type-spec-list>
    """
    match = re.compile(r'callprotoargument\b', re.I).match

    def process_item(self):
        self.specs = self.item.apply_map(self.item.get_line()[17:].lstrip())
        return

    def tofortran(self, isfix=None):
        return self.get_indent_tab(isfix=isfix) + 'CALLPROTOARGUMENT ' \
            + self.specs

# Non-standard statements


class Pause(Statement):
    """
    PAUSE [ <char-literal-constant|int-literal-constant> ]
    """
    match = re.compile(r'pause\s*(\d+|\'\w*\'|"\w*"|)\Z', re.I).match

    def process_item(self):
        self.value = self.item.apply_map(self.item.get_line()[5:].lstrip())
        return

    def tofortran(self, isfix=None):
        if self.value:
            return self.get_indent_tab(isfix=isfix) + 'PAUSE ' + self.value
        return self.get_indent_tab(isfix=isfix) + 'PAUSE'

    def analyze(self):
        return


class Comment(Statement):
    """

    Attributes
    ----------
    content : str
      Content of the comment.
    is_blank : bool
      When True then Comment represents blank line.
    """
    match = lambda s: True

    def process_item(self):
        assert self.item.comment.count('\n') <= 1, repr(self.item)
        stripped = self.item.comment.lstrip()
        self.is_blank = not stripped
        self.content = stripped[1:] if stripped else ''

    def tofortran(self, isfix=None):
        if self.is_blank:
            return ''
        if isfix:
            tab = 'C' + self.get_indent_tab(isfix=isfix)[1:]
        else:
            tab = self.get_indent_tab(isfix=isfix) + '!'
        return tab + self.content

    def analyze(self):
        return



# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# Original author: Pearu Peterson <pearu@cens.ioc.ee>
# First version created: May 2006
# Modified by J. Henrichs, Bureau of Meteorology <joerg.henrichs@bom.gov.au>

"""
Test parsing single Fortran lines.

"""

from fparser.common.splitline import splitparen, splitquote, string_replace_map


def test_splitparen():
    ''' Unit tests for splitparen function.'''
    assert splitparen('abc') == ['abc']
    assert splitparen('abc(1)') == ['abc', '(1)']
    assert splitparen('abc(1) xyz') == ['abc', '(1)', ' xyz']
    assert splitparen('a(b) = b(x,y(1)) b((a))') == \
        ['a', '(b)', ' = b', '(x,y(1))', ' b', '((a))']
    # pylint: disable=anomalous-backslash-in-string
    assert splitparen('a(b) = b(x,y(1)) b\((a)\)') == \
        ['a', '(b)', ' = b', '(x,y(1))', ' b\\(', '(a)', '\\)']
    # pylint: enable=anomalous-backslash-in-string
    assert splitparen('abc[1]') == ['abc', '[1]']
    assert splitparen('abc[1,2,3]') == ['abc', '[1,2,3]']
    assert splitparen('a[b] = b[x,y(1)] b((a))') == \
        ['a', '[b]', ' = b', '[x,y(1)]', ' b', '((a))']
    # pylint: disable=anomalous-backslash-in-string
    assert splitparen('a[b] = b[x,y(1)] b\((a)\)') == \
        ['a', '[b]', ' = b', '[x,y(1)]', ' b\\(', '(a)', '\\)']
    # pylint: enable=anomalous-backslash-in-string
    assert splitparen('integer a(3) = (/"a", "b", "c"/)') == \
        ['integer a', '(3)', ' = ', '(/"a", "b", "c"/)']
    assert splitparen(
        'character(len=40) :: a(3) = (/"a[),", ",b,[(", "c,][)("/)') == \
        ['character', '(len=40)', ' :: a', '(3)', ' = ',
         '(/"a[),", ",b,[(", "c,][)("/)']
    assert splitparen('integer a(3) = ["a", "b", "c"]') == \
        ['integer a', '(3)', ' = ', '["a", "b", "c"]']
    assert splitparen(
        'character(len=40) :: a(3) = ["a[),", ",b,[(", "c,][)("]') == \
        ['character', '(len=40)', ' :: a', '(3)', ' = ',
         '["a[),", ",b,[(", "c,][)("]']
    # pylint: disable=anomalous-backslash-in-string
    result = splitparen('a(1),b\\((2,3),c\\\\((1)),c"("')
    expected = ['a', '(1)', ',b\\(', '(2,3)', ',c\\\\', '((1))', ',c"("']
    # pylint: enable=anomalous-backslash-in-string
    assert result == expected
    # Useful for debugging:
    # for i in range(len(EXPECTED)):
    #     print i,l[i],EXPECTED[i],l[i]==EXPECTED[i]


def test_splitquote():
    '''Tests splitquote function.'''
    split_list, stopchar = splitquote('abc\\\' def"12\\"3""56"dfad\'a d\'')
    assert split_list == ['abc\\\' def', '"12\\"3"', '"56"', 'dfad', '\'a d\'']
    assert stopchar is None
    result, stopchar = splitquote('abc\\\' def"12\\"3""56"dfad\'a d\'')
    assert result == ['abc\\\' def', '"12\\"3"', '"56"', 'dfad', '\'a d\'']
    assert stopchar is None

    split_list, stopchar = splitquote('a\'')
    assert split_list == ['a', '\'']
    assert stopchar == '\''

    split_list, stopchar = splitquote('a\'b')
    assert split_list == ['a', '\'b']
    assert stopchar == '\''


def test_string_replace_map():
    '''Tests string_replace_map function.'''
    result, string_map = string_replace_map('a()')
    assert result == 'a()'
    assert string_map == {}



# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test C99 Preprocessor directives: This file tests the preprocessor
directives defined in the C99 standard. Whilst this is not part of any
Fortran standard, most compilers support preprocessing of Fortran files
and support the majority of directives from the C-Standard. Since this
is actively used in user codes there are cases where users might like to
keep the directives in the Fortran parse tree and output it again.

'''
# Author: Balthasar Reuter <balthasar.reuter@ecmwf.int>
# Based on previous work by Martin Schlipf (https://github.com/martin-schlipf)
# First version created: Jan 2020

import pytest
from fparser.two.C99Preprocessor import (
    Cpp_If_Stmt, Cpp_Elif_Stmt, Cpp_Else_Stmt, Cpp_Endif_Stmt,
    Cpp_Include_Stmt, Cpp_Macro_Stmt, Cpp_Macro_Identifier,
    Cpp_Macro_Identifier_List, Cpp_Undef_Stmt, Cpp_Line_Stmt, Cpp_Error_Stmt,
    Cpp_Warning_Stmt, Cpp_Null_Stmt, Cpp_Pp_Tokens)
from fparser.two.utils import NoMatchError
from fparser.api import get_reader


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['ABC', 'A>5', '!defined(ABC)'])
def test_pp_tokens(line):
    '''Test that the generic Cpp_Pp_Tokens class parses correctly'''
    result = Cpp_Pp_Tokens(line)
    assert str(result) == line


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [None, '', '    '])
def test_invalid_pp_tokens(line):
    '''Test that the invalid Cpp_Pp_Tokens raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Pp_Tokens(line)
    assert "Cpp_Pp_Tokens: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [
    '#if CONSTANT', '#ifdef MACRO', '#ifndef _MACRO', '#if defined(__MACRO__)',
    '#if defined __MACRO__', '#if !defined(__MACRO__)', '#if !defined MACRO'])
def test_if_stmt(line):
    '''Test that various forms of #if, #ifdef, #ifndef are recognized.'''
    result = Cpp_If_Stmt(line)
    assert str(result) == line


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line, ref', [
    ('  #  if    CONSTANT  ', '#if CONSTANT'),
    ('  #  ifdef  MACRO  ', '#ifdef MACRO'),
    ('  #  ifndef  _MACRO  ', '#ifndef _MACRO'),
    ('# if  defined(__MACRO__)    ', '#if defined(__MACRO__)'),
    ('# if  defined __MACRO__    ', '#if defined __MACRO__'),
    ('# if    !defined(__MACRO__)    ', '#if !defined(__MACRO__)'),
    ('# if  !defined __MACRO__    ', '#if !defined __MACRO__')
])
def test_if_stmt_with_whitespaces(line, ref):
    '''Test that various forms of #if, #ifdef, #ifndef are recognized when
    whitespaces are added.'''
    result = Cpp_If_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [
    None, '', ' ', '#ifdfe', '#if', '#ifdef', '#ifdef two macros'])
def test_incorrect_if_stmt(line):
    '''Test that incorrectly formed #if statements raise exception.'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_If_Stmt(line)
    assert "Cpp_If_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#elif CONDITION', '  #  elif   CONDITION '])
def test_elif_stmt(line):
    '''Test that #elif is correctly recognized.'''
    ref = '#elif CONDITION'
    result = Cpp_Elif_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [None, '', ' ', '#elfi', '#elif'])
def test_incorrect_elif_stmt(line):
    '''Test that incorrectly formed #elif statements raise exception.'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Elif_Stmt(line)
    assert "Cpp_Elif_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#else', '  # else  '])
def test_else_stmt(line):
    '''Test that #else is correctly recognized'''
    result = Cpp_Else_Stmt(line)
    assert str(result) == line


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [None, '', ' ', '#esle', '#elseA', '#Aelse'])
def test_incorrect_else_stmt(line):
    '''Test that incorrectly formed #else statements raise exception.'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Else_Stmt(line)
    assert "Cpp_Else_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#endif', '  #  endif  '])
def test_endif_stmt(line):
    '''Test that #endif is correctly recognized.'''
    result = Cpp_Endif_Stmt(line)
    assert str(result) == line


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line',
                         [None, '', ' ', '#ednif', '#endifA', '#Aendif'])
def test_incorrect_endif_stmt(line):
    '''Test that incorrectly formed #endif statements raise exception.'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Endif_Stmt(line)
    assert "Cpp_Endif_Stmt: '{0}'".format(line) in str(excinfo.value)


def test_parse_define_outside_subroutine(f2003_parser):
    '''Test parsing of #define outside of a subroutine.'''
    code = '#define MACRO\nSUBROUTINE FOO\n  CALL sub\nEND SUBROUTINE FOO'
    reader = get_reader(code)
    result = f2003_parser(reader)
    assert str(result) == code


@pytest.mark.parametrize('lines', [
    ['SUBROUTINE FOO', '  #ifdef __BAR__', '  #endif', 'END SUBROUTINE FOO'],
    ['SUBROUTINE FOO(BAR)', '  INTEGER, INTENT(INOUT) :: BAR',
     '  CALL FOOBAR(BAR)', '  #define MACRO_NAME', 'END SUBROUTINE FOO'],
    ['SUBROUTINE FOO(BAZ)', '  INTEGER, INTENT(INOUT) :: BAZ',
     '  #define MACRO_NAME', '  CALL FOOBAZ(BAR)', 'END SUBROUTINE FOO'],
    ['SUBROUTINE FOO(BAZ2)', '  #define MACR', '  INTEGER, INTENT(IN) :: BAZ2',
     '  CALL FOOBAZ2(BAR)', 'END SUBROUTINE FOO'],
    ['SUBROUTINE FOO(BAR2)', '  INTEGER, INTENT(IN) :: BAR2', '  #ifdef MACRO',
     '  CALL FOO2(BAR)', '  #else', '  CALL ABORT("error msg")', '  #endif',
     '  BAR2 = BAR2 + 1', 'END SUBROUTINE FOO'],
    ['SUBROUTINE FOO(BAR3)', '  #ifdef MACRO', '  INTEGER, INTENT(IN) :: BAR3',
     '  CALL FOOBAR3(BAR)', '  #endif', 'END SUBROUTINE FOO'],
    ['#define MACRO', 'SUBROUTINE BAR', '  #if defined(MACRO)', '  CALL sub1',
     '  #elif FOO', '  CALL sub2', '  #else', '  CALL sub3', '  #endif',
     'END SUBROUTINE BAR']
])
def test_directive_at_different_places(f2003_parser, lines):
    '''Test that directives are parsed correctly irrespective of where they
    appear in the code.'''
    code = '\n'.join(line.strip() for line in lines)
    ref = '\n'.join(lines)
    reader = get_reader(code)
    result = f2003_parser(reader)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#include "filename.inc"',
                                  '   #   include  "filename.inc"  '])
def test_include_stmt_quotes(line):
    '''Test that #include is recognized'''
    ref = '#include "filename.inc"'
    result = Cpp_Include_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#include <filename.inc>',
                                  '   #   include  <filename.inc>  '])
def test_include_stmt_bracket(line):
    '''Test that #include is recognized'''
    ref = '#include "filename.inc"'
    result = Cpp_Include_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [
    None, '', '  ', '#includ', '#includ "x"', '#include', '#include ""',
    "#include 'x'", '#include "x', '#include x"', '#include x',
    '#include x"x"', '#include "x"x', 'x #include "x"', '#includex "x"',
    "#include 'abc'", '#include " a.inc"', '#include "      "'])
def test_incorrect_include_stmt(line):
    '''Test that incorrectly formed #include statements raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Include_Stmt(line)
    assert "Cpp_Include_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [
    '#define MACRO value', '#define _MACRO', '#define MACRO(x) call func(x)',
    '#define MACRO (x, y)', '#define MACRO(x, y) (x) + (y)',
    '#define MACRO(a, b, c, d) (a) * (b) + (c) * (d)',
    '#define eprintf(...) fprintf (stderr, __VA_ARGS__)',
    '#define report(tst, ...) ((tst)?puts(#tst):printf(__VA_ARGS__))',
    '#define hash_hash # ## #', '#define TABSIZE 100', '#define r(x,y) x ## y',
    '#define MACRO(a, bbb, c_d) (a) * (bbb + c_d)', '#define MACRO x',
    '#define MACRO( a,b2_ ,   c) (a )*    (   b2_   + c  )',
    '#define omp_get_num_threads() 1', '#define MACRO(a2aa, b, c)'])
def test_macro_stmt(line):
    '''Test that #define is recognized'''
    result = Cpp_Macro_Stmt(line)
    assert str(result) == line


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line, ref', [
    ('  #  define   MACRO   value  ', '#define MACRO value'),
    ('   #  define  _MACRO  ', '#define _MACRO')])
def test_macro_stmt_with_whitespace(line, ref):
    '''Test that #define is recognized when white space are added.'''
    result = Cpp_Macro_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [
    None, '', ' ', '#def', '#defnie', '#definex', '#define 2a'
    '#define fail(...,test) test', '#define', '#define fail(...,...)'])
def test_incorrect_macro_stmt(line):
    '''Test that incorrectly formed #define statements raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Macro_Stmt(line)
    assert "Cpp_Macro_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('name', [
    'MACRO', 'MACRO1', 'MACRO_', 'MACRO_NAME', 'macro', '_', '_12'])
def test_macro_identifier(name):
    '''Test that all allowed names can be parsed'''
    result = Cpp_Macro_Identifier(name)
    assert str(result) == name


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('name',
                         ['12MACRO', '+MACRO1', '//MACRO_', 'MAC RO', ''])
def test_invalid_macro_identifier(name):
    '''Test that invalid names raise exceptions'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Macro_Identifier(name)
    assert "Cpp_Macro_Identifier: '{0}'".format(name) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['(a, b, ...)', '(...)', '(a)', '(a, b)'])
def test_macro_identifier_list(line):
    '''Test that correct lists are parsed'''
    result = Cpp_Macro_Identifier_List(line)
    assert str(result) == line


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line',
                         [None, '(a, ..., b)', '(..., ...)', '(..)', '(....)'])
def test_invalid_macro_identifier_list(line):
    '''Test that invalid lists raise exceptions'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Macro_Identifier_List(line)
    msg = "Cpp_Macro_Identifier_List: '{0}'".format(line)
    assert msg in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#undef _MACRO', '   #  undef  _MACRO  '])
def test_undef_stmt(line):
    '''Test that #undef is recognized'''
    ref = '#undef _MACRO'
    result = Cpp_Undef_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [
    None, '', ' ', '#undef', '#unfed', '#undefx'])
def test_incorrect_undef_stmt(line):
    '''Test that incorrectly formed #undef statements raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Undef_Stmt(line)
    assert "Cpp_Undef_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#line CONFIG', '  #  line   CONFIG  '])
def test_line_statement(line):
    '''Test that #line is recognized'''
    ref = '#line CONFIG'
    result = Cpp_Line_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [None, '', ' ', '#line', '#linex', '#lien'])
def test_incorrect_line_stmt(line):
    '''Test that incorrectly formed #line statements raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Line_Stmt(line)
    assert "Cpp_Line_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#error MSG', '  #  error  MSG  '])
def test_error_statement_with_msg(line):
    '''Test that #error is recognized'''
    # error with message
    ref = '#error MSG'
    result = Cpp_Error_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#error', '  #   error'])
def test_error_statement_without_msg(line):
    '''Test that #error is recognized'''
    # error without message
    ref = '#error'
    result = Cpp_Error_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [None, '', ' ', '#erorr', '#errorx'])
def test_incorrect_error_stmt(line):
    '''Test that incorrectly formed #error statements raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Error_Stmt(line)
    assert "Cpp_Error_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#warning MSG', '  #  warning  MSG  '])
def test_warning_statement_with_msg(line):
    '''Test that #warning is recognized (not actually part of C99)
    with a message.'''
    ref = '#warning MSG'
    result = Cpp_Warning_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#warning ', '  #  warning'])
def test_warning_statement_without_msg(line):
    '''Test that #warning is recognized (not actually part of C99)
    without a message.'''
    ref = '#warning'
    result = Cpp_Warning_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', [None, '', ' ', '#wrning', '#warningx'])
def test_incorrect_warning_stmt(line):
    '''Test that incorrectly formed #warning statements raise exception'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Warning_Stmt(line)
    assert "Cpp_Warning_Stmt: '{0}'".format(line) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line', ['#', '   #  ', '#   ', '   #'])
def test_null_stmt(line):
    '''Test that null directives are recognized'''
    ref = '#'
    result = Cpp_Null_Stmt(line)
    assert str(result) == ref


@pytest.mark.usefixtures("f2003_create")
@pytest.mark.parametrize('line',
                         [None, '', ' ', '#a', '##', '###', '  # #', '# a'])
def test_incorrect_null_stmt(line):
    '''Test that anything that is not a single hash sign is not recognized'''
    with pytest.raises(NoMatchError) as excinfo:
        _ = Cpp_Null_Stmt(line)
    assert "Cpp_Null_Stmt: '{0}'".format(line) in str(excinfo.value)



# Modified work Copyright (c) 2017-2019 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Various utility functions.

-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

__all__ = ['split_comma', 'specs_split_comma',
           'ParseError','AnalyzeError',
           'get_module_file','parse_bind','parse_result','is_name','parse_array_spec',
           'CHAR_BIT','str2stmt',
           'classes']

import logging
import glob
import io
import os
import re
import traceback

import six


class ParseError(Exception):
    pass

class AnalyzeError(Exception):
    pass

is_name = re.compile(r'^[a-z_]\w*$',re.I).match
name_re = re.compile(r'[a-z_]\w*',re.I).match
is_entity_decl = re.compile(r'^[a-z_]\w*',re.I).match
is_int_literal_constant = re.compile(r'^\d+(_\w+|)$').match
module_file_extensions = ['.f', '.f90', '.f95', '.f03', '.f08']


def split_comma(line, item=None, comma=',', keep_empty=False,
                brackets=None):
    ''' Split (an optionally bracketed) comma-separated list into
    items and return a list containing them. If supplied then
    brackets must be a list of containing two strings, the first
    being the opening bracket and the second the closing bracket. '''
    # we may have blank space so strip the line
    line = line.strip()
    if not line:
        return []
    if brackets:
        if not isinstance(brackets, tuple):
            raise ParseError("split_comma: brackets must be a tuple")
        if len(brackets) != 2:
            raise ParseError("split_comma: brackets tuple must contain "
                             "just two items but got: {0}", brackets)
        open = brackets[0]
        close = brackets[1]
        if not line.startswith(open) or not line.endswith(close):
            return []
        line = line.strip(brackets[0])
        line = line.strip(brackets[1])
    items = []
    if item is None:
        for s in line.split(comma):
            s = s.strip()
            if not s and not keep_empty: continue
            items.append(s)
        return items
    newitem = item.copy(line, True)
    apply_map = newitem.apply_map
    for s in newitem.get_line().split(comma):
        s = apply_map(s).strip()
        if not s and not keep_empty: continue
        items.append(s)
    return items


def extract_bracketed_list_items(line, item=None):
    ''' Takes any line that contains "xxx (a,b,...) yyy" and returns
    a list of items corresponding to a, b, ... Anything outside of
    the parentheses is ignored. Only works for strings containing
    a single set of parentheses. '''
    if line.count('(') > 1 or line.count(')') > 1:
        raise ParseError(
            "parse_bracketed_list: more than one opening/closing parenthesis "
            "found in string '{0}'; this is not supported".format(line))
    idx1 = line.find('(')
    idx2 = line.rfind(')')
    if idx1 < 0 or idx2 < 0 or idx2 < idx1:
        raise ParseError(
            "parse_bracketed_list: failed to find expression within "
            "parentheses in '{0}'".format(line))
    items = split_comma(line[idx1:idx2+1], item,
                        brackets=("(", ")"))
    if item:
        for idx in range(len(items)):
            itm = item.copy(items[idx])
            rlst = []
            for rpart in itm.get_line().split(':'):
                rlst.append(itm.apply_map(rpart.strip()))
            items[idx] = rlst
    return items


def parse_array_spec(line, item = None):
    items = []
    for spec in split_comma(line, item):
        items.append(tuple(split_comma(spec, item, comma=':', keep_empty=True)))
    return items

def specs_split_comma(line, item = None, upper=False):
    specs0 = split_comma(line, item)
    specs = []
    for spec in specs0:
        i = spec.find('=')
        if i!=-1:
            kw = spec[:i].strip().upper()
            v  = spec[i+1:].strip()
            specs.append('%s = %s' % (kw, v))
        else:
            if upper:
                spec = spec.upper()
            specs.append(spec)
    return specs

def parse_bind(line, item=None):
    if not line.lower().startswith('bind'):
        return None, line
    if item is not None:
        newitem = item.copy(line, apply_map=True)
        newline = newitem.get_line()
    else:
        newitem = None
    newline = newline[4:].lstrip()
    i = newline.find(')')
    assert i != -1, repr(newline)
    args = []
    for a in specs_split_comma(newline[1:i].strip(), newitem, upper=True):
        args.append(a)
    rest = newline[i+1:].lstrip()
    if item is not None:
        rest = newitem.apply_map(rest)
    return args, rest

def parse_result(line, item=None):
    if not line.lower().startswith('result'):
        return None, line
    line = line[6:].lstrip()
    i = line.find(')')
    assert i != -1, repr(line)
    name = line[1:i].strip()
    assert is_name(name), repr(name)
    return name, line[i+1:].lstrip()

def filter_stmts(content, classes):
    """ Pop and return classes instances from content.
    """
    stmts = []
    indices = []
    for i in range(len(content)):
        stmt = content[i]
        if isinstance(stmt, classes):
            stmts.append(stmt)
            indices.append(i)
    indices.reverse()
    for i in indices:
        del content[i]
    return stmts


def get_module_files(directory, _cache={}):
    if directory in _cache:
        return _cache[directory]
    module_line = re.compile(r'(\A|^)module\s+(?P<name>\w+)\s*(!.*|)$',re.I | re.M)
    d = {}
    files = []
    for ext in module_file_extensions:
        files += glob.glob(os.path.join(directory,'*'+ext))
    for fn in files:
        f = open(fn,'r')
        for name in module_line.findall(f.read()):
            name = name[1]
            if name in d:
                print(d[name],'already defines',name)
                continue
            d[name] = fn
    _cache[directory] = d
    return d

def get_module_file(name, directory, _cache={}):
    fn = _cache.get(name, None)
    if fn is not None:
        return fn
    if name.endswith('_module'):
        for ext in module_file_extensions:
            f1 = os.path.join(directory,name[:-7]+ext)
            if os.path.isfile(f1):
                _cache[name] = fn
                return f1
    files = []
    for ext in module_file_extensions:
        files += glob.glob(os.path.join(directory,'*'+ext))
    for fn in files:
        if module_in_file(name, fn):
            _cache[name] = fn
            return fn
    return

def module_in_file(name, filename):
    name = name.lower()
    pattern = re.compile(r'\s*module\s+(?P<name>[a-z]\w*)', re.I).match
    f = open(filename,'r')
    for line in f:
        m = pattern(line)
        if m and m.group('name').lower()==name:
            f.close()
            return filename
    f.close()


def str2stmt(string, isfree=True, isstrict=False):
    """ Convert Fortran code to Statement tree.
    """
    from .readfortran import Line, FortranStringReader
    from .parsefortran import FortranParser
    reader = FortranStringReader(string, isfree, isstrict)
    parser = FortranParser(reader)
    parser.parse()
    parser.analyze()
    block = parser.block
    while len(block.content)==1:
        block = block.content[0]
    return block

def show_item_on_failure(func, _exception_depth=[0]):
    """
    Decorator for analyze methods.
    """
    def new_func(self):
        try:
            func(self)
        except AnalyzeError as msg:
            clsname = self.__class__.__name__
            self.error('%s.analyze error: %s' % (clsname,msg))
            traceback.print_exc()
        except ParseError as msg:
            self.error('parse error: %s' % (msg))
        except Exception as msg:
            _exception_depth[0] += 1
            if _exception_depth[0]==1:
                self.error('exception triggered here: %s %s' % (Exception, msg))
            raise
        _exception_depth[0] = 0
    return new_func

_classes_cache = {}
class meta_classes(type):
    """ Meta class for ``classes``.
    """
    __abstractmethods__ = False
    def __getattr__(self, name):
        # Expose created classes only as attributes to ``classes`` type.
        cls = _classes_cache.get(name)
        if cls is None:
            raise AttributeError('instance does not have attribute %r' % (name))
        return cls


class classes(six.with_metaclass(meta_classes, type)):
    """Make classes available as attributes of this class.

    To add a class to the attributes list, one must use::

      __metaclass__ = classes

    in the definition of the class.

    In addition, apply the following tasks:

    * decorate analyze methods with show_item_on_failure
    """

    def __new__(metacls, name, bases, dict):
        if 'analyze' in dict:
            dict['analyze'] =  show_item_on_failure(dict['analyze'])
        cls = type.__new__(metacls, name, bases, dict)
        _classes_cache[name] = cls
        return cls


def make_clean_tmpfile(filename, skip_bad_input=True, encoding="utf8"):
    '''An input file may contain invalid characters which, in Python3
    causes an exception when the file is read. By default, this utility
    removes any invalid characters from the input file, writes the
    cleaned version into a temporary file, and returns the name of the
    newly created temporary file. It is up to the user of this
    function to remove the temporary file once it is no longer
    required. If the skip_bad_input optional argument is set to
    'False' then an exception will be raised if invalid characters are
    found in either Python2 or 3.

    :param str filename: the name of the original filename
    :param bool skip_bad_input: Optional argument specifying whether \
    to ignore and remove invalid input ('True') or whether to raise an \
    exception ('False'). Defaults to 'True'.
    :param str encoding: Optional argument specifying the encoding to \
    use when reading the input file. Defaults to 'ascii'.

    :returns: the name of the temporary file created by this function.
    :rtype: str

    :raises InternalError: if the skip_bad_input argument has an \
    invalid (not False or True)
    :raises ParseError: if invalid input is found in the input file \
    and the argument 'skip_bad_input' is set to 'False'.

    '''
    import tempfile
    import codecs
    from fparser.two.utils import InternalError

    if skip_bad_input not in [False, True]:
        raise InternalError(
            "utils.py: make_clean_tmpfile: skip_bad_input argument should "
            "be False or True but found '{0}'.".format(skip_bad_input))

    try:
        orig_file = codecs.open(filename, "r", encoding=encoding)
        _ = orig_file.read()
    except LookupError as excinfo:
        raise InternalError(excinfo)
    except UnicodeDecodeError as excinfo:
        message = ("character in input file. Error returned was "
                   "{0}.".format(str(excinfo)))
        if not skip_bad_input:
            raise ParseError("Bad "+message)
        # Log the fact that this character will be removed from the input file
        logging.getLogger(__name__).warning("Skipped bad %s", message)
    orig_file.close()

    # Tell codec to skip any errors
    with io.open(filename, "r",
                 encoding=encoding, errors='ignore') as orig_file:
        file_input = orig_file.read()

    if six.PY2:
        # Unicode needs to be encoded.
        file_input = file_input.encode('UTF-8')

    encoding = {'encoding': 'UTF-8'} if six.PY3 else {}
    # Set delete to False so file will not be deleted when closed.
    with tempfile.NamedTemporaryFile(mode='w',
                                     delete=False, **encoding) as temp_file:
        temp_file.write(file_input)

    return temp_file.name



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''A simple fparser2 Fortran2008 example demonstrating support for
submodules'''
from fparser.two.parser import ParserFactory
from fparser.common.readfortran import FortranStringReader
MYFILE = '''
program hello
integer a
end program hello
submodule (hello2) world
end submodule world
subroutine world2
end subroutine world2
'''
READER = FortranStringReader(MYFILE)
F2008_PARSER = ParserFactory().create(std="f2008")
PROGRAM = F2008_PARSER(READER)
print(PROGRAM)



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 Cray-pointers: This file tests the support for a
Cray-pointer declaration.

'''

import pytest
from fparser.two.Fortran2003 import Cray_Pointer_Decl
from fparser.two.utils import NoMatchError, InternalError


def test_cray_pointer_decl(f2003_create):
    '''Check that Cray-pointer declarations are parsed correctly.

    '''
    for myinput in ["(a, b)", "  ( a , b )  "]:
        ast = Cray_Pointer_Decl(myinput)
        assert "(a, b)" in str(ast)
        assert repr(ast) == ("Cray_Pointer_Decl(Name('a'), Name('b'))")


def test_pointee_decl(f2003_create):
    '''Check that a Cray-pointer declaration containing a pointee
    declaration is parsed correctly (for explicit and assumed shape
    arrays).

    '''
    for myinput in ["(a, b(n))", "(a, b(0 : n))", "(a, b(n, m))",
                    "(a, b(5, *))", "(a, b(*))", "(a, b(0 : 1, 2 : *))"]:
        ast = Cray_Pointer_Decl(myinput)
        assert myinput in str(ast)


def test_errors(f2003_create):
    '''Check that syntax errors produce a NoMatchError exception.'''
    for myinput in [None, "", "  ", "a, b)", "(a, b", "()", "(a)", "(a b)",
                    "(1, a)", "(a, 1)", "(a, b(2)", "(a, b2))", "(a, b())"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Cray_Pointer_Decl(myinput)
        assert "Cray_Pointer_Decl: '{0}'".format(myinput) in str(excinfo.value)


def test_internal_error1(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the length of the Items
    list is not 2 as the str() method assumes that it is.

    '''
    myinput = "(mypointer, mypointee)"
    ast = Cray_Pointer_Decl(myinput)
    monkeypatch.setattr(ast, "items", [None])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "should be of size 2 but found '1'" in str(excinfo.value)


def test_internal_error2(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the pointer name (entry 0
    of Items) is empty or None as the str() method assumes that it is
    a string with content.

    '''
    myinput = "(mypointer, mypointee)"
    ast = Cray_Pointer_Decl(myinput)
    for change in [None, ""]:
        monkeypatch.setattr(ast, "items", (change, "mypointee"))
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("'Items' entry 0 should be a pointer name but it is "
                "empty") in str(excinfo.value)


def test_internal_error3(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the pointee name (entry 1
    of Items) is empty or None as the str() method assumes that it is
    a string with content.

    '''
    myinput = "(mypointer, mypointee)"
    ast = Cray_Pointer_Decl(myinput)
    for change in [None, ""]:
        monkeypatch.setattr(ast, "items", ("mypointer", change))
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("'Items' entry 1 should be a pointee name or pointee "
                "declaration but it is empty") in str(excinfo.value)
